/*
 * qsockreader.cpp
 *
 *  Created on: Aug 1, 2013
 *      Author: root
 */

#include "../Includes/qstdinc.h"
#include "../Includes/qsock.h"
#include "../Includes/qlogger.h"
#include "../Includes/qsockreader.h"

QLOG_REGISTER_SOURCE

/* Default params */
static QReadSockParams g_defaults;

/* Reader function */
bool QReadSocketLine(QBuffer& buffer, std::string& sLine, QTCPSock* pSock, QReadSockParams* pInParams)
{
	/* Obtain valid pointers */
	QReadSockParams* pParams = pInParams?pInParams:&g_defaults;

	/* Try to read from socket */
	bool bYeild = true;
	bool bRetVal = false;
	QByteArray holder;
	holder.resize(pParams->nOneRead);

	int nInputSize = 0;
	if (!pSock->GetInputSize(&nInputSize)) return false;
	if (nInputSize > 0)
	{
		/* Bounds check / receive */
		if (nInputSize > pParams->nOneRead) nInputSize = pParams->nOneRead;
		if (!pSock->GenRecv(holder.data(), nInputSize)) return false;
		bYeild = false;

		/* Analyze data */
		buffer.Add(holder.data(), nInputSize);
	}
	else
	{
		if (!pSock->CheckConnect()) return false;
	}

	/* Clear line */
	sLine.clear();

	/* Analyze buffer */
	if ((int)buffer.GetSize() > pParams->nLineMaxLen)
	{
		if (pParams->bDebug)
		{
			QNetAddr rAddr;
			pSock->GetPeer(&rAddr);

			QLogInfo(0, "HTTP line length from source (%s) is too long, closing connection", rAddr.GetAddrIP() );
		}
		buffer.Clean();
		pSock->Close();
		return false;
	}

	/* Analyze buffer for complete line */
	uint8_t* pData = buffer.GetData();
	int nDataSize = (int)buffer.GetSize();
	for (int i=0; i<(nDataSize - 1); i++)
	{
		/* Check for CRLF */
		if ((pData[i] == 0x0d) && (pData[i + 1] == 0x0a))
		{
			/* Found CRLF, extract to string and remove from buffer */
			sLine = std::string((const char*)pData, i);
			buffer.Remove(0, i + 2, NULL);
			if (pParams->bDebug)
			{
				QLogDebug(0, "--> New line (%s)", sLine.c_str());
			}
			bRetVal = true;
			break;
		}
	}

	/* Sleep a little if required */
	if (pParams->nWaitTime > 0)
	{
		if ((sLine.length() == 0) && bYeild)
		{
#ifndef _WIN32
			usleep(100000);
#else
			::Sleep(100);
#endif
		}
	}

	return bRetVal;
}

bool QReadBufferLine(QBuffer& buffer, std::string& sLine, uint8_t* pInData, int nInSize, QReadSockParams* pInParams)
{
	/* Obtain valid pointers */
	QReadSockParams* pParams = pInParams?pInParams:&g_defaults;

	/* Analyze data */
	if (nInSize)
	{
		buffer.Add(pInData, nInSize);
	}

	bool bRetVal = false;

	/* Clear line */
	sLine.clear();

	/* Analyze buffer */
	if ((int)buffer.GetSize() > pParams->nLineMaxLen)
	{
		if (pParams->bDebug)
		{
			QLogInfo(0, "Input line length from is too long" );
			buffer.Clean();
		}
		return false;
	}

	/* Analyze buffer for complete line */
	uint8_t* pData = buffer.GetData();
	int nDataSize = (int)buffer.GetSize();
	for (int i=0; i<(nDataSize - 1); i++)
	{
		/* Check for CRLF */
		if ((pData[i] == 0x0d) && (pData[i + 1] == 0x0a))
		{
			/* Found CRLF, extract to string and remove from buffer */
			sLine = std::string((const char*)pData, i);
			buffer.Remove(0, i + 2, NULL);
			if (pParams->bDebug)
			{
				QLogDebug(0, "--> New line (%s)(%d)", sLine.c_str(), buffer.GetSize());
			}
			bRetVal = true;
			break;
		}
	}

	return bRetVal;
}

int QURLCharToNumber(char c)
{
	switch(c)
	{
		case '0':
			return 0;
		case '1':
			return 1;
		case '2':
			return 2;
		case '3':
			return 3;
		case '4':
			return '4';
		case '5':
			return 5;
		case '6':
			return 6;
		case '7':
			return 7;
		case '8':
			return 8;
		case '9':
			return 9;
		case 'a':
		case 'A':
			return 10;
		case 'b':
		case 'B':
			return 11;
		case 'c':
		case 'C':
			return 12;
		case 'd':
		case 'D':
			return 13;
		case 'e':
		case 'E':
			return 14;
		case 'f':
		case 'F':
			return 15;
	}
	return -1;
}

bool QURLIsHex(char c)
{
	return QURLCharToNumber(c) >= 0;
}

bool QURLDecode(std::string& sLine)
{
	std::string sDst;
	int nSrcLen = (int)sLine.length();

	for (int i=0; i<nSrcLen;)
	{
		if ((sLine[i] == '%') && (i < (nSrcLen - 2)))
		{
			/* Probably escape sequence */
			if (QURLIsHex(sLine[i + 1]) && QURLIsHex(sLine[i + 2]))
			{
				int nChar = (QURLCharToNumber(sLine[i + 1]) << 4) | QURLCharToNumber(sLine[i + 2]);
				sDst = sDst + (char)nChar;
			}
			else
			{
				sDst = sDst + sLine[i];
				sDst = sDst + sLine[i + 1];
				sDst = sDst + sLine[i + 2];
			}
			i += 3;
		}
		else
		{
			sDst = sDst + sLine[i];
			i++;
		}
	}

	/* Copy back */
	sLine = sDst;
	return true;
}
